/**
   \file ccfv.h
   \author Haniel Barbosa
   \brief Module for searching conflicting instances

   Given a set of groundly consistent literals L and a quantified formula
   Ax1..xn.F a conflicting instantiation \sigma is a substitution s.t. L \wedge
   F\sigma is unsatisfiable. */
#ifndef __CCFV_H
#define __CCFV_H

#include "unify.h"
#include "ujobs.h"

#ifdef PROOF
#include "proof.h"
#endif

/**
   [TODO] Ugly workaround to have clause as well */
typedef struct TDAGinst
{
  TDAG qform;
  Tstack_DAG clause;
  Tstack_Tunifier insts;
} TDAGinst;

TSstack(_DAGinst, TDAGinst); /* typedefs Tstack_DAGinsts */

/*
  --------------------------------------------------------------
  IO interface
  --------------------------------------------------------------
*/

/**
   \author Haniel Barbosa
   \brief initializes module */
extern void CCFV_init(void);

/**
   \author Haniel Barbosa
   \brief releases module */
extern void CCFV_done(void);

extern void CCFV_cycle_init(void);
extern void CCFV_cycle_done(void);

/**
   \author Haniel Barbosa
   \brief checks all quantified formulas (inedependently) and builds
   conflicting instantiations
   \return all instances created along with the quantified formula/clause they
   instantiate */
extern Tstack_DAGinst CCFV(void);

/*
  --------------------------------------------------------------
  Workaround for triggers
  --------------------------------------------------------------
*/

extern void match_DAGs(Tstack_Tunifier * result, TDAG NGDAG, Tstack_DAG terms);

extern
void combine_unifiers(Tstack_Tunifier *result, Tstack_Tunifier base,
                      Tstack_Tunifier new);

#endif
