/**
   \file proof.h
   \author Pascal Fontaine

   \brief proof module.

   This module provides API functions to memorize the proofs done in
   veriT. */

#ifndef __PROOF_H
#define __PROOF_H

#include "config.h"

#ifdef PROOF

#include <stdarg.h>
#include <stdio.h>

#include "bool.h"
#include "DAG.h"
#include "stack.h"

#include "proof-sat-solver.h"
#include "proof-rules-tautologies.h"
#include "proof-rules.h"
#include "proof-step-table.h"
#include "proof-subproof.h"
#include "proof-output.h"
#include "proof-lemma-hash.h"
#include "proof-unsat-core.h"

extern bool proof_on;

extern char * option_proof_filename;
extern bool option_proof_file_from_input;
extern bool option_proof_stat;

extern bool proof_no_replacement;

extern bool proof_with_sharing;
extern bool option_proof_prune;
extern bool option_proof_merge;

/**
   \addtogroup arguments_user

   - --proof-version

   defines proof output format (default is 0); 1 is old one, using ordered sets
   with references; 2 is older one, using resolution clauses */
extern int proof_version;

/**
   \addtogroup arguments_user

   - --proof-stats

   outputs statistics about proofs */
extern bool proof_stats;

/*
  --------------------------------------------------------------
  Init/Done
  --------------------------------------------------------------
*/

/**
   \author Pascal Fontaine
   \brief module initialisation */
extern void proof_init(void);

/**
   \author Pascal Fontaine
   \brief module release */
extern void proof_done(void);

/**
   \author Pascal Fontaine
   \brief outputs proof documentation to file */
extern void proof_doc(FILE * file);

/**
   \author Pascal Fontaine
   \brief notifies module of the input file name */
extern void proof_set_input_file(char * filename);

#endif /* PROOF */

#ifdef PEDANTIC
void      proof_done(void);
#endif

#endif /* __PROOF_H */
