#define _BSD_SOURCE /* for mkdtemp */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "general.h"
#include "sys.h"

FILE *
sys_open_file(const char * name, const char * mode)
{
  FILE * result;
  result = fopen(name, mode);
  if (result == 0)
    {
      my_error("cannot open file \"%s\".\n", name);
      exit(SYS_EXIT_ERROR_IO);
    }
  return result;
}

/*--------------------------------------------------------------*/

void
sys_close_file(FILE * file)
{
  if (fclose(file) == -1)
    {
      my_error("cannot close file.\n");
      exit(SYS_EXIT_ERROR_IO);
    }
}

/*--------------------------------------------------------------*/

char *
sys_create_directory(void)
{
  /** TEMPLATE is a string that specifies how directory and files
   * created by veriT shall be named. See mkdtemp (3). */

  static const char TEMPLATE [] = "veriT-XXXXXX";

  char * template;
  MY_MALLOC(template, sizeof(char) * (strlen(TEMPLATE) + 1));
  strcpy(template, TEMPLATE);
#ifndef WIN32
  return mkdtemp(template);
#else
  /* This is insecure, but mkdtemp is not available on Windows. */
  mktemp(template);
  mkdir(template);
  return template;
#endif
}
